import logging
from functools import singledispatchmethod

from dataclass_factory import Factory
from sikei.container.protocol import Container
from sikei.events.event import DomainEvent, ECSTEvent, Event, NotificationEvent
from sikei.events.map import EventMap
from sikei.message_brokers.protocol import Message, MessageBroker

logger = logging.getLogger(__name__)


class EventEmitter:
    """
    The event emitter is responsible for sending events to the according handlers or
    to the message broker abstraction.

    Usage::

      redis_client = Redis()  # async redis client
      message_broker = RedisMessageBroker(redis_client)
      event_map =  EventMap()
      event_map.bind(UserJoinedDomainEvent, UserJoinedDomainEventHandler)

      event_emitter = EventEmitter(event_map, container, message_broker)

      # Uses UserJoinedDomainEventHandler for handling event:
      await event_emitter.emit(user_joined_domain_event)

      # Sends event to the Redis Pub/Sub:
      await event_emitter.emit(user_joined_notification_event)

    """

    def __init__(
        self,
        event_map: EventMap,
        container: Container,
        message_broker: MessageBroker | None = None,
    ) -> None:
        self._event_map = event_map
        self._container = container
        self._message_broker = message_broker

    @singledispatchmethod
    async def emit(self, event: Event) -> None:
        ...

    @emit.register
    async def _(self, event: DomainEvent) -> None:
        handlers_types = self._event_map.get(type(event))

        for handler_type in handlers_types:
            handler = await self._container.resolve(handler_type)
            logger.debug(
                "Handling Event(%s) via event handler(%s)",
                type(event).__name__,
                handler_type.__name__,
            )
            await handler.handle(event)

    @emit.register
    async def _(self, event: NotificationEvent) -> None:
        if not self._message_broker:
            raise RuntimeError("To use NotificationEvent, message_broker argument must be specified.")

        message = _build_message(event)

        logger.debug(
            "Sending Notification Event(%s) to message broker %s",
            event.event_id,
            type(self._message_broker).__name__,
        )

        await self._message_broker.send_message(message)

    @emit.register
    async def _(self, event: ECSTEvent) -> None:
        if not self._message_broker:
            raise RuntimeError("To use ECSTEvent, message_broker argument must be specified.")

        message = _build_message(event)

        logger.debug(
            "Sending ECST event(%s) to message broker %s",
            event.event_id,
            type(self._message_broker).__name__,
        )

        await self._message_broker.send_message(message)


def _build_message(event: NotificationEvent | ECSTEvent) -> Message:
    factory = Factory()

    payload = factory.dump(event)

    return Message(
        message_type=event._event_type,
        message_name=type(event).__name__,
        message_id=event.event_id,
        payload=payload,
    )



from typing import Type

from sikei.requests.request import Request
from sikei.requests.request_handler import RequestHandler


class RequestMap:
    def __init__(self) -> None:
        self._request_map: dict[Type[Request], Type[RequestHandler]] = {}

    def bind(
        self,
        request_type: Type[Request],
        handler_type: Type[RequestHandler],
    ) -> None:
        self._request_map[request_type] = handler_type

    def get(self, request_type: Type[Request]) -> Type[RequestHandler]:
        handler_type = self._request_map.get(request_type)
        if not handler_type:
            raise RequestHandlerDoesNotExist("RequestHandler not found matching Request type.")

        return handler_type

    def __str__(self) -> str:
        return str(self._request_map)


class RequestHandlerDoesNotExist(Exception):
    ...



from dataclasses import dataclass, field

from sikei.events.event import Event
from sikei.response import Response


@dataclass
class DispatchResult:
    response: Response | None = field(default=None)
    events: list[Event] = field(default_factory=list)



from typing import Type, TypeVar

import di
from di.dependent import Dependent
from di.executors import AsyncExecutor
from sikei.container.protocol import Container

T = TypeVar("T")


class DIContainer(Container[di.Container]):
    def __init__(self) -> None:
        self._external_container: di.Container | None = None

    @property
    def external_container(self) -> di.Container:
        if not self._external_container:
            raise AttributeError

        return self._external_container

    def attach_external_container(self, container: di.Container) -> None:
        self._external_container = container

    async def resolve(self, type_: Type[T]) -> T:
        executor = AsyncExecutor()
        solved = self.external_container.solve(Dependent(type_, scope="request"), scopes=["request"])
        with self.external_container.enter_scope("request") as state:
            return await solved.execute_async(executor=executor, state=state)



from typing import Protocol, Type, TypeVar

T = TypeVar("T")
C = TypeVar("C")


class Container(Protocol[C]):
    """
    The container interface.
    """

    @property
    def external_container(self) -> C:
        ...

    def attach_external_container(self, container: C) -> None:
        ...

    async def resolve(self, type_: Type[T]) -> T:
        ...



import orjson
from azure.servicebus import ServiceBusMessage
from azure.servicebus.aio import ServiceBusClient
from sikei.message_brokers.protocol import Message


class AzureMessageBroker:
    def __init__(self, client: ServiceBusClient, topic_name: str, *, timeout: float | None = None) -> None:
        self._client = client
        self._topic_name = topic_name
        self._timeout = timeout

    async def send_message(self, message: Message) -> None:
        async with self._client:
            sender = self._client.get_topic_sender(self._topic_name)

            service_bus_message = _parse_message(message)

            await sender.send_messages(service_bus_message, timeout=self._timeout)


def _parse_message(message: Message) -> ServiceBusMessage:
    return ServiceBusMessage(orjson.dumps(message), content_type="application/json")




import aio_pika
import orjson
from sikei.message_brokers.protocol import Message


class RabbitMQMessageBroker:
    def __init__(self, client, *, routing_key: str | None = None, exchange: str | None = None) -> None:
        self._connection = client
        self._routing_key = routing_key 
        self._exchange = exchange 

    async def send_message(self, message: Message) -> None:
        async with self._connection as connection:
            
            channel = await connection.channel()
            if self._exchange:
                await channel.basic_publish(exchange=self._exchange,body=aio_pika.Message(body=orjson.dumps(message)),routing_key=self._routing_key)
            else:
                await channel.default_exchange.publish(aio_pika.Message(body=orjson.dumps(message)),routing_key=self._routing_key)



import functools
from typing import Awaitable, Callable, Protocol, TypeVar

from sikei.requests import Request
from sikei.response import Response

Req = TypeVar("Req", bound=Request, contravariant=True)
Res = TypeVar("Res", Response, None, covariant=True)
HandleType = Callable[[Req], Awaitable[Res]]


class Middleware(Protocol):
    async def __call__(self, request: Request, handle: HandleType) -> Res:
        ...


Handle = Callable[[Req], Awaitable[Res]]


class MiddlewareChain:
    def __init__(self) -> None:
        self._chain: list[Middleware] = []

    def set(self, chain: list[Middleware]) -> None:
        self._chain = chain

    def add(self, middleware: Middleware) -> None:
        self._chain.append(middleware)

    def wrap(self, handle: Handle) -> Handle:
        for middleware in reversed(self._chain):
            handle = functools.partial(middleware.__call__, handle=handle)

        return handle



from dataclasses import dataclass


@dataclass(frozen=True, kw_only=True)
class Response:
    """
    Base class for response type objects.

    The response is a result of the request handling, which hold by RequestHandler.

    Often the response is used for defining the result of the query.

    Usage::

        @dataclass(frozen=True, kw_only=True)
        class ReadMeetingQueryResult(Response):
            meeting_id: int = field()
            link: str = field()
            status: MeetingStatusEnum = field()

    """



from sikei.container.protocol import Container
from sikei.dispatcher.dispatch_result import DispatchResult
from sikei.middlewares.base import MiddlewareChain
from sikei.requests.map import RequestMap
from sikei.requests.request import Request


class DefaultDispatcher:
    def __init__(
        self,
        request_map: RequestMap,
        container: Container,
        middleware_chain: MiddlewareChain | None = None,
    ) -> None:
        self._request_map = request_map
        self._container = container
        self._middleware_chain = middleware_chain or MiddlewareChain()

    async def dispatch(self, request: Request) -> DispatchResult:
        handler_type = self._request_map.get(type(request))

        handler = await self._container.resolve(handler_type)

        wrapped_handle = self._middleware_chain.wrap(handler.handle)

        response = await wrapped_handle(request)

        return DispatchResult(response=response, events=handler.events)



from sikei.events.event import DomainEvent, ECSTEvent, Event, NotificationEvent
from sikei.events.event_emitter import EventEmitter
from sikei.events.event_handler import EventHandler
from sikei.events.map import EventMap

__all__ = (
    "Event",
    "DomainEvent",
    "ECSTEvent",
    "NotificationEvent",
    "EventEmitter",
    "EventHandler",
    "EventMap",
)



from collections import defaultdict
from typing import Type, TypeVar

from sikei.events.event import DomainEvent
from sikei.events.event_handler import EventHandler

E = TypeVar("E", bound=DomainEvent, contravariant=True)


class EventMap:
    def __init__(self) -> None:
        self._event_map: dict[Type[DomainEvent], list[Type[EventHandler]]] = defaultdict(lambda: [])

    def bind(self, event_type: Type[E], handler_type: Type[EventHandler[E]]) -> None:
        self._event_map[event_type].append(handler_type)

    def get(self, event_type: Type[E]) -> list[Type[EventHandler[E]]]:
        return self._event_map[event_type]

    def get_events(self) -> list[Type[DomainEvent]]:
        return list(self._event_map.keys())

    def __str__(self) -> str:
        return str(self._event_map)



from typing import Type

from sikei.container.protocol import Container
from sikei.dispatcher import DefaultDispatcher, Dispatcher
from sikei.events import Event, EventEmitter
from sikei.middlewares import MiddlewareChain
from sikei.requests import Request, RequestMap
from sikei.response import Response


class Mesikei:
    """
    The main mediator object.

    Usage::

      redis_client = Redis()  # async redis client
      message_broker = RedisMessageBroker(redis_client)
      event_map = EventMap()
      event_map.bind(UserJoinedDomainEvent, UserJoinedDomainEventHandler)
      request_map = RequestMap()
      request_map.bind(JoinUserCommand, JoinUserCommandHandler)
      event_emitter = EventEmitter(event_map, container, message_broker)

      mediator = Mesikei(
        event_emitter=event_emitter,
        request_map=request_map,
        container=container
      )

      # Handles command and published events by the command handler.
      await mediator.send(join_user_command)

    """

    def __init__(
        self,
        request_map: RequestMap,
        container: Container,
        event_emitter: EventEmitter | None = None,
        middleware_chain: MiddlewareChain | None = None,
        *,
        dispatcher_type: Type[Dispatcher] = DefaultDispatcher,
    ) -> None:
        self._event_emitter = event_emitter
        self._dispatcher = dispatcher_type(
            request_map=request_map, container=container, middleware_chain=middleware_chain  # type: ignore
        )

    async def send(self, request: Request) -> Response | None:
        dispatch_result = await self._dispatcher.dispatch(request)

        if dispatch_result.events:
            await self._send_events(dispatch_result.events.copy())

        return dispatch_result.response

    async def _send_events(self, events: list[Event]) -> None:
        if not self._event_emitter:
            return

        while events:
            event = events.pop()
            await self._event_emitter.emit(event)



from dataclasses import dataclass, field
from datetime import datetime
from uuid import UUID, uuid4


@dataclass(frozen=True, kw_only=True)
class Event:
    ...


@dataclass(frozen=True, kw_only=True)
class DomainEvent(Event):
    """
    The base class for domain events.
    """


@dataclass(frozen=True, kw_only=True)
class NotificationEvent(Event):
    """
    The base class for notification events.

    Contains only identification information about state change.

    Example plain structure::

      {
          "event_id": "82a0b10e-1b3d-4c3c-9bdd-3934f8f824c2",
          "event_timestamp": "2023-03-06 12:11:35.103792",
          "changed_user_id": 987
      }

    """

    event_id: UUID = field(default_factory=uuid4)
    event_timestamp: datetime = field(default_factory=datetime.utcnow)
    _event_type = "notification_event"


@dataclass(frozen=True, kw_only=True)
class ECSTEvent(Event):
    """
    Base class for ECST events.

    ECST means event-carried state transfer.

    Contains full information about state change.

    Example plain structure::

      {
          "event_id": "82a0b10e-1b3d-4c3c-9bdd-3934f8f824c2",
          "event_timestamp": "2023-03-06 12:11:35.103792",
          "user_id": 987,
          "new_user_last_name": "Doe",
          "new_user_nickname": "kend"
      }

    """

    event_id: UUID = field(default_factory=uuid4)
    event_timestamp: datetime = field(default_factory=datetime.utcnow)
    _event_type = "ecst_event"



from typing import Protocol, TypeVar

from sikei.events.event import DomainEvent

E = TypeVar("E", bound=DomainEvent, contravariant=True)


class EventHandler(Protocol[E]):
    """
    The event handler interface.

    Usage::

      class UserJoinedEventHandler(EventHandler[UserJoinedEventHandler])
          def __init__(self, meetings_api: MeetingAPIProtocol) -> None:
              self._meetings_api = meetings_api

          async def handle(self, event: UserJoinedEventHandler) -> None:
              await self._meetings_api.notify_room(event.meeting_id, "New user joined!")

    """

    async def handle(self, event: E) -> None:
        raise NotImplementedError



import logging
from typing import Awaitable, Callable, Mapping, Protocol, TypeVar

from sikei.requests import Request
from sikei.response import Response

Req = TypeVar("Req", bound=Request, contravariant=True)
Res = TypeVar("Res", Response, None, covariant=True)
HandleType = Callable[[Req], Awaitable[Res]]


class Logger(Protocol):
    def log(self, level: int, msg: str, *args, extra: Mapping[str, object] | None = None) -> None:  # noqa
        ...


class LoggingMiddleware:
    def __init__(
        self,
        logger: Logger | None = None,
        level: int = logging.DEBUG,
    ) -> None:
        self._logger = logger or logging.getLogger(__name__)
        self._level = level

    async def __call__(self, request: Request, handle: HandleType) -> Res:
        self._logger.log(
            self._level,
            "Handle %s request",
            type(request).__name__,
            extra={"request": request},
        )
        response = await handle(request)
        self._logger.log(
            self._level,
            "Request %s handled. Response: %s",
            type(request).__name__,
            response,
            extra={"request": request},
        )

        return response



import logging

import orjson
from redis.asyncio import Redis
from sikei.message_brokers.protocol import Message

logger = logging.getLogger(__name__)


class RedisMessageBroker:
    def __init__(self, client: Redis, *, channel_prefix: str | None = None) -> None:
        self._client = client
        self._channel_prefix = channel_prefix or "python_sikei_channel"

    async def send_message(self, message: Message) -> None:
        async with self._client.pubsub() as pubsub:
            channel = f"{self._channel_prefix}:{message.message_type}:{message.message_id}"

            await pubsub.subscribe(channel)

            logger.debug("Sending message to Redis Pub/Sub %s.", message.message_id)
            await self._client.publish(channel, orjson.dumps(message))



from typing import Type, TypeVar

import rodi
from sikei.container.protocol import Container

T = TypeVar("T")


class RodiContainer(Container[rodi.Container]):
    def __init__(self) -> None:
        self._external_container: rodi.Container | None = None

    @property
    def external_container(self) -> rodi.Container:
        if not self._external_container:
            raise AttributeError

        return self._external_container

    def attach_external_container(self, container: rodi.Container) -> None:
        self._external_container = container

    async def resolve(self, type_: Type[T]) -> T:
        if hasattr(self.external_container, "resolve"):
            return self.external_container.resolve(type_)
        return self._build_by_provider(type_)

    def _build_by_provider(self, type_: Type[T]) -> T:
        services = self.external_container.build_provider()
        return services.get(type_)



from dataclasses import dataclass, field
from typing import Protocol
from uuid import UUID, uuid4


@dataclass(frozen=True, kw_only=True)
class Message:
    message_type: str = field()
    message_name: str = field()
    message_id: UUID = field(default_factory=uuid4)
    payload: dict = field()


class MessageBroker(Protocol):
    """
    The interface over a message broker.

    Used for sending messages to message brokers (currently only redis supported).
    """

    async def send_message(self, message: Message) -> None:
        ...



from typing import Type, TypeVar

from dependency_injector import containers

T = TypeVar("T")


class DependencyInjectorContainer:
    def __init__(self) -> None:
        self._external_container: containers.DeclarativeContainer | None = None

    @property
    def external_container(self) -> containers.DeclarativeContainer:
        if not self._external_container:
            raise AttributeError("External container is not attached.")

        return self._external_container

    def attach_external_container(self, container: containers.DeclarativeContainer) -> None:
        self._external_container = container

    async def resolve(self, type_: Type[T]) -> T:
        provider = self.external_container.providers['dependency']

        return provider()



from typing import Protocol, TypeVar

from sikei.events.event import Event
from sikei.requests.request import Request
from sikei.response import Response

Req = TypeVar("Req", bound=Request, contravariant=True)
Res = TypeVar("Res", Response, None, covariant=True)


class RequestHandler(Protocol[Req, Res]):
    """
    The request handler interface.

    The request handler is an object, which gets a request as input and may return a response as a result.

    Command handler example::

      class JoinMeetingCommandHandler(RequestHandler[JoinMeetingCommand, None])
          def __init__(self, meetings_api: MeetingAPIProtocol) -> None:
              self._meetings_api = meetings_api
              self.events: list[Event] = []

          async def handle(self, request: JoinMeetingCommand) -> None:
              await self._meetings_api.join_user(request.user_id, request.meeting_id)

    Query handler example::

      class ReadMeetingQueryHandler(RequestHandler[ReadMeetingQuery, ReadMeetingQueryResult])
          def __init__(self, meetings_api: MeetingAPIProtocol) -> None:
              self._meetings_api = meetings_api
              self.events: list[Event] = []

          async def handle(self, request: ReadMeetingQuery) -> ReadMeetingQueryResult:
              link = await self._meetings_api.get_link(request.meeting_id)
              return ReadMeetingQueryResult(link=link, meeting_id=request.meeting_id)

    """

    @property
    def events(self) -> list[Event]:
        ...

    async def handle(self, request: Req) -> Res:
        raise NotImplementedError



from sikei.dispatcher.default import DefaultDispatcher
from sikei.dispatcher.dispatch_result import DispatchResult
from sikei.dispatcher.protocol import Dispatcher

__all__ = (
    "DispatchResult",
    "DefaultDispatcher",
    "Dispatcher",
)



from typing import Protocol

from sikei.dispatcher.dispatch_result import DispatchResult
from sikei.requests.request import Request


class Dispatcher(Protocol):
    async def dispatch(self, request: Request) -> DispatchResult:
        ...



from dataclasses import dataclass, field
from uuid import UUID, uuid4


@dataclass(frozen=True, kw_only=True)
class Request:
    """
    Base class for request-type objects.

    The request is an input of the request handler.
    Often Request is used for defining queries or commands.

    Usage::

      @dataclass(frozen=True, kw_only=True)
      class JoinMeetingCommand(Request):
          meeting_id: int = field()
          user_id: int = field()

      @dataclass(frozen=True, kw_only=True)
      class ReadMeetingByIdQuery(Request):
          meeting_id: int = field()

    """

    request_id: UUID = field(default_factory=uuid4)
